.transparent-bg {
    background-color: transparent;
    background-image: none;
}

.thumbnail {
    background: @theme_base_color;
    color: @theme_fg_color;
    border: 1px solid @borders;
}

.live-thumbnail {
    background: black;
}

/* This replaces a separator between the recommended
 * OS list and the buttons at the bottom. */
.index-stack {
    border-bottom: 1px solid @borders;
}

.boxes-product-key-entry {
    font-family: monospace;
}

/* This is to be removed once we port to AdwToast */
.boxes-toast {
     border-radius: 2em;
}

/* Screenshot flash */
@keyframes flash_animation {
    50% {
        border: 100px solid #FFF;
        opacity: 0.2;
    }
}

.flash {
    border: 1px solid transparent;
    animation: flash_animation;
    animation-duration: 150ms;
    animation-iteration-count: 1;
}

.overlay-box {
    padding: 25px;
    border-radius: 5px;
}

/* Welcome Tutorial */
.tutorial-page {
    background-repeat: no-repeat;
    background-size: 60%;
    background-position: center 15%;
}

.welcome-tutorial,
.welcome-tutorial-title-label,
.welcome-tutorial-description-label {
    color: white;
}
